/*
 * bootstrap.c -- This program loads the Linux/68k kernel into an Amiga
 *                and and launches it.
 *
 * Copyright (C) 1994 by Waldorf GMBH,
 * written by Ralf Baechle
 *
 * Loosly based on bootstrap.c for Linux/68k
 *
 * Copyright (C) 1993 by Hamish Macdonald, Greg Harp
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file README.legal in the main directory of this archive
 * for more details.
 */

#include <stddef.h>
#include <arc/stdlib.h>
#include <sys/file.h>
#include <sys/types.h>
#include <unistd.h>
#include <asm/system.h>
#include <arc/pathname.h>
#include <linux/string.h>


/*
 * Defaults, may be overiden by option or environment.
 */
static char *kernel_name = "multi(0)disk(0)fdisk(0)\\vmlinux";
static char *ramdisk_name = NULL;

/*
 * MIPS bootstrap include file
 */
#include "bootstrap.h"

/*
 * required Linux/MIPS include files
 */
#include <linux/a.out.h>
#include "bootinfo.h"

extern char *optarg;

extern volatile void launch(char *memptr, u_long kernel_size, u_long rd_size);
struct exec kexec;
char *memptr;
u_long start_mem = 0x80000000;		/* kseg0 */
u_long mem_size;
u_long rd_size;
u_long kernel_entry;
u_long kernel_bss_end;

struct bootinfo bi;

void usage(void)
{
	fprintf (stderr, "Usage:\n"
		 "\tbootstrap [-d] [-k kernel_executable] [-r ramdisk_file]"
		 " [option...]\n");
	exit (EXIT_FAILURE);
}

int main(int argc, char *argv[])
{
    int ch, debugflag = 1, kfd, rfd = -1, i;
    long fast_total = 0;	     /* total Fast RAM in system */
    struct MemHeader *mnp;
    struct ConfigDev *cdp = NULL;
    u_long memreq;
    struct nlist *nl;
    long startcodesize;
    u_long *stack, kbi_offset;

    /* print the greet message */
    printf("Linux/MIPS ARC Bootstrap version 0.0\n\r");
	
    /* check arguments */
    while ((ch = getopt(argc, argv, "dk:r:")) != EOF)
        switch (ch) {
            case 'd':
                debugflag = 1;
                break;
            case 'k':
                kernel_name = optarg;
                break;
            case 'r':
                ramdisk_name = optarg;
                break;
            case '?':
            default:
                usage();
            }
	argc -= optind;
	argv += optind;

	/*
	 * FIXME: should use the ARC BIOS
	 */

	bi.machtype   = MACH_DESKSTATION_TYNE;
	bi.cputype    = CPU_R4600;
	bi.memlower   = 0x80000000;
	bi.memupper   = 0x80800000;
	bi.sec_cache  = 512*1024;
	bi.dma_cache  = 512*1024;

	/*
	 * tell us where the kernel will go
	 */

	printf("\nThe kernel will be located at %08lx\n\r", start_mem);
	/*
	 * verify that there is enough RAM
	 */
	if(bi.memupper - bi.memlower < 0x800000) {
	    printf("Insufficient Memory to load Linux/MIPS, aborting\n\r");
	    exit(5);
	    }
    /*
     * open kernel executable and read exec header
     */
    if ((kfd = open (kernel_name, O_RDONLY)) == -1) {
        printf ("Unable to open kernel file %s\r\n", kernel_name);
        exit (EXIT_FAILURE);
        }
    if (read (kfd, (void *)&kexec, sizeof(kexec)) != sizeof(kexec)) {
        printf ("Unable to read exec header from %s\n\r", kernel_name);
        exit (EXIT_FAILURE);
        }
    if (ramdisk_name) {
        if ((rfd = open (ramdisk_name, O_RDONLY)) == -1) {
            fprintf (stderr, "Unable to open ramdisk file %s\n\r", ramdisk_name);
            exit (EXIT_FAILURE);
            }
        /*
         * record ramdisk size
         */
        bi.ramdisk_size = (lseek (rfd, 0, L_XTND) + 1023) >> 10;
        }
    else
        bi.ramdisk_size = 0;
	
    rd_size = bi.ramdisk_size << 10;
    bi.ramdisk_addr = (u_long)start_mem + mem_size - rd_size;

    /*
     * find offset to boot_info structure
     */
    if (!(nl = get_nlist (kernel_name, "_boot_info"))) {
        perror ("get_nlist");
        exit (EXIT_FAILURE);
        }
    else {
        kbi_offset = nl->n_value - kexec.a_entry;
        free (nl);
        }
    /*
     * Find kernel entry point
     */
    if (!(nl = get_nlist (kernel_name, "kernel_entry"))) {
        perror ("get_nlist");
        exit (EXIT_FAILURE);
        }
    else {
        kernel_entry = nl->n_value;
        free (nl);
        }

    /*
     * End of bss segment - ramdisk will be placed there
     */
    kernel_bss_end = 0x80000000 + kexec.a_text + kexec.a_data + kexec.a_bss;
	
    memreq = kexec.a_text + kexec.a_data + rd_size;
    if (!(memptr = (char *) malloc(memreq))) {
        fprintf (stderr, "Unable to allocate memory\n\r");
        exit (EXIT_FAILURE);
        }
    memset(memptr, 0, memreq);
	
    if (lseek (kfd, N_TXTOFF(kexec), L_SET) == -1) {
        fprintf (stderr, "Failed to seek to text\n\r");
        free(memptr);
        exit (EXIT_FAILURE);
        }
    if (read (kfd, memptr, kexec.a_text) != kexec.a_text) {
        fprintf (stderr, "Failed to read text\n\r");
        free(memptr);
        exit (EXIT_FAILURE);
        }
    if (lseek (kfd, N_DATOFF(kexec), L_SET) == -1) {
        fprintf (stderr, "Failed to seek to data\n\r");
        free(memptr);
        exit (EXIT_FAILURE);
        }
    if (read (kfd, memptr + kexec.a_text, kexec.a_data) != kexec.a_data) {
        fprintf (stderr, "Failed to read data\n\r");
        free(memptr);
        exit (EXIT_FAILURE);
	}
    close (kfd);
    /*
     * copy the boot_info struct to the kernel image
     */
    memcpy ((void *)(memptr + kbi_offset), &bi, sizeof(bi));
	
    if (rfd != -1) {
        if (lseek (rfd, 0, L_SET) == -1) {
            fprintf (stderr, "Failed to seek to beginning of ramdisk file\n\r");
            free(memptr);
            exit (EXIT_FAILURE);
            }
        if (read(rfd, memptr+kexec.a_text+kexec.a_data, rd_size) != rd_size) {
            fprintf (stderr, "Failed to read ramdisk file\n\r");
            free(memptr);
            exit (EXIT_FAILURE);
            }
	close (rfd);
        }
    /*
     * Should check whether Kernel & RAMdisk moving doesn't overwrite
     * the bootstraper during startup.
     */
    if (debugflag) {
        if (bi.ramdisk_size)
            printf ("RAM disk at %#lx, size is %ldK\n\r",
                    (u_long)memptr + kexec.a_text + kexec.a_data,
                    bi.ramdisk_size);
	  
        printf ("\nKernel text at %#lx, code size %d\n\r",
		  start_mem + N_TXTADDR(kexec), kexec.a_text);
        printf ("Kernel data at %#lx, data size %d\n\r",
		  start_mem + N_DATADDR(kexec), kexec.a_data );
        printf ("Kernel bss  at %#lx, bss  size %d\n\r",
		  start_mem + N_BSSADDR(kexec), kexec.a_bss );
	  
        printf ("\nKernel entry is %#x\n\r", kexec.a_entry );
	  
        printf ("ramdisk dest top is %#lx\n\r", start_mem + mem_size);
        printf ("ramdisk lower limit is %#lx\n\r",
                (u_long)(memptr + kexec.a_text + kexec.a_data));
        printf ("ramdisk src top is %#lx\n\r",
                (u_long)(memptr + kexec.a_text + kexec.a_data)
                + rd_size);
	  
        printf ("Type a key to continue the Linux boot...");
        fflush (stdout);
        getchar();
        printf ("\r\n");
        }
	
    /*
     * wait for things to settle down
     */
    sleep(2);

    /*
     * execute the copy-and-go code
     */

#if 1
    launch(memptr, kexec.a_text + kexec.a_data, rd_size);
#else
    memcpy((PVOID)0x80000000, memptr, kexec.a_text + kexec.a_data);

    FlushAllCaches();

    __asm__ __volatile__(
	"la	$4,0x80000000\n\t"
	"jr	$4");
#endif

    /* NOTREACHED */

    return(0);
}
